# -*- coding: utf-8 -*-

"""
.. module:: pyessv._model.authority.py
   :copyright: Copyright "December 01, 2016", IPSL
   :license: GPL/CeCIL
   :platform: Unix, Windows
   :synopsis: A vocabulary authority, e.g. WGCM.

.. moduleauthor:: Mark Conway-Greenslade <momipsl@ipsl.jussieu.fr>


"""
import pyessv
from pyessv._model.entity import Entity



class Authority(Entity):
    """An authority assuming responsibity for governance of vocabularies.

    """
    def __init__(self):
        """Instance constructor.

        """
        super(Authority, self).__init__(pyessv.ENTITY_TYPE_AUTHORITY)

        self.scopes = list()


    def __len__(self):
        """Returns number of items in managed collection.

        """
        return len(self.scopes)


    def __iter__(self):
        """Instance iterator initializer.

        """
        return iter(sorted(self.scopes,
                           key=lambda i: i if isinstance(i, basestring) else i.name))


    def __getitem__(self, key):
        """Returns a child section item.

        """
        return Entity.get_item(self, key)


    def __contains__(self, key):
        """Instance membership predicate.

        """
        return self[key] is not None


    @property
    def ancestors(self):
        """Gets ancestors within archive hierarchy.

        """
        return []


    @property
    def hierarchy(self):
        """Gets hierachy within archive.

        """
        return [self]


    @property
    def namespace(self):
        """Returns namespace used in I/O scenarios.

        """
        return self.name


    def parse(self, scope_name, strict=True):
        """Parses an associated scope name.

        """
        return pyessv.parse(scope_name, self, strict=strict)


    def accept(self):
        """Marks entity as accepted.

        """
        for scope in self:
            scope.accept()


    def deprecate(self):
        """Marks entity as deprecated.

        """
        for scope in self:
            scope.deprecate()


    def destroy(self):
        """Marks entity for removal from all persistant state stores.

        """
        for scope in self:
            scope.destroy()


    def reject(self):
        """Marks entity as rejected.

        """
        for scope in self:
            scope.reject()


    def reset(self):
        """Resets entity status.

        """
        for scope in self:
            scope.reset()



# -*- coding: utf-8 -*-

"""
.. module:: pyessv._factory.py
   :copyright: Copyright "December 01, 2016", IPSL
   :license: GPL/CeCIL
   :platform: Unix, Windows
   :synopsis: Encapsulates creation of domain model class instances.

.. moduleauthor:: Mark Conway-Greenslade <momipsl@ipsl.jussieu.fr>

"""
import uuid

import arrow

import pyessv

from pyessv._constants import ENTITY_TYPE_AUTHORITY
from pyessv._constants import ENTITY_TYPE_COLLECTION
from pyessv._constants import ENTITY_TYPE_SCOPE
from pyessv._constants import ENTITY_TYPE_TERM
from pyessv._exceptions import ValidationError
from pyessv._model import ENTITY_TYPE_KEY_MAP
from pyessv._model import Authority
from pyessv._model import Entity
from pyessv._model import Scope
from pyessv._model import Collection
from pyessv._model import Term
from pyessv._utils.compat import basestring
from pyessv._utils.compat import str
from pyessv._validation import validate_entity
from pyessv._parser_template import TemplateParser



def create_authority(name, description=None, label=None, url=None, create_date=None, data=None):
    """Instantiates, initialises & returns a term authority.

    :param str name: Canonical authority name.
    :param str description: Authority description.
    :param str label: Label for UI purposes.
    :param str url: Authority further information URL.
    :param datetime create_date: Date upon which authority was created.
    :param dict data: Arbirtrary data associated with authority.

    :returns: A vocabulary authority, e.g. wcrp.
    :rtype: pyessv.Authority

    """
    instance = _create_entity(Authority, name, description, label, url, create_date, data)
    errors = validate_entity(instance)
    if errors:
        raise ValidationError(errors)

    return instance


def create_scope(authority, name, description=None, label=None, url=None, create_date=None, data=None):
    """Instantiates, initialises & returns a term scope.

    :param pyessv.Authority authority: CV authority to which scope is bound.
    :param str name: Canonical scope name.
    :param str label: Label for UI purposes.
    :param str description: Scope description.
    :param str url: Scope URL for further information.
    :param datetime create_date: Date upon which scope was created.
    :param dict data: Arbirtrary data associated with scope.

    :returns: A vocabulary scope, e.g. cmip6.
    :rtype: pyessv.Scope

    """
    instance = _create_entity(Scope, name, description, label, url, create_date, data, authority)
    instance.authority = authority
    errors = validate_entity(instance)
    if errors:
        raise ValidationError(errors)

    return instance


def create_collection(scope, name, description=None, label=None, url=None, create_date=None, data=None, term_name_regex=None):
    """Instantiates, initialises & returns a term collection.

    :param pyessv.Scope scope: CV scope to which collection is bound.
    :param str name: Canonical collection name.
    :param str description: Collection description.
    :param str label: Label for UI purposes.
    :param str url: Collection URL for further information.
    :param datetime create_date: Date upon which collection was created.
    :param dict data: Arbirtrary data associated with collection.

    :returns: A vocabulary collection, e.g. insitution-id.
    :rtype: pyessv.Collection

    """
    instance = _create_entity(Collection, name, description, label, url, create_date, data, scope)
    instance.scope = scope
    instance.term_name_regex = term_name_regex
    errors = validate_entity(instance)
    if errors:
        raise ValidationError(errors)

    return instance


def create_term(collection, name, description=None, label=None, url=None, create_date=None, data=None):
    """Instantiates, initialises & returns a term.

    :param pyessv.Collection collection: The collection to which the term belongs.
    :param str name: Canonical term name.
    :param str description: Term description.
    :param str label: Label for UI purposes.
    :param str url: Term URL for further information.
    :param datetime create_date: Date upon which term was created.
    :param dict data: Arbitrary data associated with term.

    :returns: A vocabulary term, e.g. ipsl.
    :rtype: pyessv.Term

    """
    instance = _create_entity(Term, name, description, label, url, create_date, data, collection)
    instance.collection = collection
    instance.idx = len(collection)
    errors = validate_entity(instance)
    if errors:
        raise ValidationError(errors)

    return instance


def create_template_parser(template, collections):
    """Instantiates, initialises & returns a template parser.

    :param str template: An expression template.
    :param tuple collections: Collections that the template maps to.

    :returns: A vocabulary expression parser.
    :rtype: pyessv.TemplateParser

    """
    assert isinstance(template, basestring), 'Invalid template'
    assert isinstance(collections, tuple), 'Invalid collections'
    assert len(template) > 0, 'Invalid template'
    assert len(collections) > 0, 'Invalid collections'
    assert len([i for i in collections if not isinstance(i, Collection)]) ==0, 'Invalid collections'
    assert template.count('{}') == len(collections), 'Invalid template'

    return TemplateParser(template, collections)


def _create_entity(typeof, name, description, label=None, url=None, create_date=None, data=None, owner=None):
    """Instantiates, initialises & returns an entity.

    """
    # Set core attributes.
    instance = typeof()
    instance.description = description
    instance.label = name
    instance.name = str(name)
    instance.create_date = create_date or arrow.utcnow().datetime
    instance.data = data
    instance.uid = uuid.uuid4()
    instance.url = url

    # Set associative attributes.
    if owner is not None:
        # ... parent in vocab hierarchy
        if isinstance(instance, Scope):
            instance.authority = owner
        elif isinstance(instance, Collection):
            instance.scope = owner
        elif isinstance(instance, Term):
            instance.collection = owner
        # ... child in vocab hierarchy
        if isinstance(owner, Authority):
            owner.scopes.append(instance)
        elif isinstance(owner, Scope):
            owner.collections.append(instance)
        elif isinstance(owner, Collection):
            owner.terms.append(instance)

    return instance



# -*- coding: utf-8 -*-

"""
.. module:: pyessv._model.term.py
   :copyright: Copyright "December 01, 2016", IPSL
   :license: GPL/CeCIL
   :platform: Unix, Windows
   :synopsis: A vocabulary term, e.g. IPSL.

.. moduleauthor:: Mark Conway-Greenslade <momipsl@ipsl.jussieu.fr>


"""
import pyessv
from pyessv._model.entity import Entity



class Term(Entity):
    """A vocabulary term.

    """
    def __init__(self):
        """Instance constructor.

        """
        super(Term, self).__init__(pyessv.ENTITY_TYPE_TERM)

        self.alternative_name = None    # primary synonym
        self.alternative_url = None     # an alternative URL
        self.associations = list()      # associated terms
        self.collection = None          # collection of which term is a member
        self.idx = None                 # collection ordinal
        self.parent = None              # parent term within collection
        self.status = pyessv.GOVERNANCE_STATUS_PENDING
        self.synonyms = list()          # name synonyms


    def __contains__(self, key):
        """Instance membership predicate.

        """
        key = str(key).strip().lower()

        return key in self.all_names


    @property
    def hierarchy(self):
        """Gets hierachy within archive.

        """
        return [self.authority, self.scope, self.collection, self]


    @property
    def ancestors(self):
        """Gets ancestors within archive hierarchy.

        """
        return [self.authority, self.scope, self.collection]


    @property
    def namespace(self):
        """Returns namespace used in I/O scenarios.

        """
        return '{}:{}'.format(self.collection.namespace, self.name)


    @property
    def owner(self):
        """Gets owner within vocabulary model.

        """
        return self.collection


    @property
    def authority(self):
        """Gets associated governing authority.

        """
        return self.scope.authority


    @property
    def scope(self):
        """Gets associated scope.

        """
        return self.collection.scope


    @property
    def all_names(self):
        """Returns all term names.

        """
        result = [self.name, self.alternative_name] + self.synonyms
        result = [t for t in result if t is not None and len(t) > 0]

        return set(sorted(result))


    @property
    def depth(self):
        """Gets hierarchical depth.

        """
        return len(self.ancestors)


    @property
    def ancestors(self):
        """Gets ancestral hierarchy.

        """
        result = []
        ancestor = self.parent
        while ancestor:
            result.append(ancestor)
            ancestor = ancestor.parent
        result.reverse()

        return result


    def add_synonym(self, new_synonym):
        """Adds a new synonym to the term's synnym set.

        """
        # Validate new synonym.
        _validate_term_synonym(new_synonym)

        # Format.
        new_synonym = str(new_synonym).strip()

        # Escape if already aliased.
        for synonym in self.synonyms:
            if synonym == new_synonym:
                return

        # Update synonym set.
        self.synonyms = sorted(self.synonyms + [new_synonym])

        # Save term.
        # self.partition.save(self)


    def save(self):
        """Saves term to persistant state stores.

        """
        self.partition.save(self)


    def associate(self, term):
        """Appends an associated term to managed collection.

        :param pyessv.Term term: Associated term to be added.

        """
        self.associations.add(term)


    def accept(self):
        """Marks entity as accepted.

        """
        self.status = pyessv.GOVERNANCE_STATUS_ACCEPTED


    def deprecate(self):
        """Marks entity as deprecated.

        """
        self.status = pyessv.GOVERNANCE_STATUS_DEPRECATED


    def destroy(self):
        """Marks entity for removal from all persistant state stores.

        """
        self.status = pyessv.GOVERNANCE_STATUS_DEPRECATED


    def reject(self):
        """Marks entity as rejected.

        """
        self.status = pyessv.GOVERNANCE_STATUS_REJECTED


    def reset(self):
        """Resets entity status.

        """
        self.status = pyessv.GOVERNANCE_STATUS_PENDING



# -*- coding: utf-8 -*-

"""
.. module:: pyessv._model.__init__.py
   :copyright: Copyright "December 01, 2016", IPSL
   :license: GPL/CeCIL
   :platform: Unix, Windows
   :synopsis: Model of vocabulary entities.

.. moduleauthor:: Mark Conway-Greenslade <momipsl@ipsl.jussieu.fr>


"""
from pyessv._constants import ENTITY_TYPE_AUTHORITY
from pyessv._constants import ENTITY_TYPE_COLLECTION
from pyessv._constants import ENTITY_TYPE_SCOPE
from pyessv._constants import ENTITY_TYPE_TERM
from pyessv._model.authority import Authority
from pyessv._model.collection import Collection
from pyessv._model.entity import Entity
from pyessv._model.scope import Scope
from pyessv._model.term import Term



# Set of supported entity types.
ENTITY_TYPES = (
    Authority,
    Collection,
    Scope,
    Term,
)

# Map of type keys to types.
ENTITY_TYPE_KEY_MAP = {
    Authority: ENTITY_TYPE_AUTHORITY,
    Collection: ENTITY_TYPE_COLLECTION,
    Scope: ENTITY_TYPE_SCOPE,
    Term: ENTITY_TYPE_TERM
}



# -*- coding: utf-8 -*-

"""
.. module:: pyessv._model.entity.py
   :copyright: Copyright "December 01, 2016", IPSL
   :license: GPL/CeCIL
   :platform: Unix, Windows
   :synopsis: An entity within the pyessv domain model.

.. moduleauthor:: Mark Conway-Greenslade <momipsl@ipsl.jussieu.fr>


"""
import uuid

import arrow

from pyessv._constants import ENTITY_TYPE_AUTHORITY
from pyessv._constants import ENTITY_TYPE_COLLECTION
from pyessv._constants import ENTITY_TYPE_SCOPE
from pyessv._constants import ENTITY_TYPE_TERM
from pyessv._utils.compat import basestring



class Entity(object):
    """An entity within the pyessv domain model.

    """
    def __init__(self, typekey):
        """Instance constructor.

        """
        self.create_date = None
        self.data = None
        self.description = None
        self.label = None
        self.name = None
        self.typekey = typekey
        self.uid = None
        self.url = None


    def __repr__(self):
        """Instance representation.

        """
        return self.namespace


    def validate(self):
        """Validates instance.

        :returns: Set of validation errrors.
        :rtype: set

        """
        # N.B. just-in-time import to avoid circular references.
        from pyessv._validation import validate_entity

        return validate_entity(self)


    @property
    def errors(self):
        """Returns set of validation errors.

        """
        return sorted(list(self.validate()))


    @property
    def is_valid(self):
        """Gets flag indicating validity.

        """
        return len(self.validate()) == 0


    @staticmethod
    def get_item(entity, key):
        """Returns an item from managed collection.

        """
        # Set collection.
        if entity.typekey == ENTITY_TYPE_AUTHORITY:
            items = entity.scopes
        elif entity.typekey == ENTITY_TYPE_COLLECTION:
            items = entity.terms
        elif entity.typekey == ENTITY_TYPE_SCOPE:
            items = entity.collections

        # Set comparator to be used.
        if isinstance(key, int):
            comparator = lambda i: i.idx
        elif isinstance(key, uuid.UUID):
            comparator = lambda i: i.uid
        else:
            key = str(key).strip().lower()
            try:
                uuid.UUID(key)
            except ValueError:
                comparator = lambda i: i.name
            else:
                comparator = lambda i: str(i.uid)

        # Match against a attribute.
        for item in items:
            if comparator(item) == key:
                return item

        # Match against a synonym.
        try:
            items = [i for i in items if i.synonyms]
        except AttributeError:
            pass
        else:
            for item in items:
                if key in item.synonyms:
                    return item
